# Copyright 2014 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import mock


class Test__LocalStack(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from google.cloud._helpers import _LocalStack

        return _LocalStack

    def _make_one(self):
        return self._get_target_class()()

    def test_it(self):
        batch1, batch2 = object(), object()
        batches = self._make_one()
        self.assertEqual(list(batches), [])
        self.assertIsNone(batches.top)
        batches.push(batch1)
        self.assertIs(batches.top, batch1)
        batches.push(batch2)
        self.assertIs(batches.top, batch2)
        popped = batches.pop()
        self.assertIs(popped, batch2)
        self.assertIs(batches.top, batch1)
        self.assertEqual(list(batches), [batch1])
        popped = batches.pop()
        self.assertIsNone(batches.top)
        self.assertEqual(list(batches), [])


class Test__UTC(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from google.cloud._helpers import _UTC

        return _UTC

    def _make_one(self):
        return self._get_target_class()()

    def test_module_property(self):
        from google.cloud import _helpers as MUT

        klass = self._get_target_class()
        try:
            import pytz
        except ImportError:  # pragma: NO COVER
            self.assertIsInstance(MUT.UTC, klass)
        else:
            self.assertIs(MUT.UTC, pytz.UTC)

    def test_dst(self):
        import datetime

        tz = self._make_one()
        self.assertEqual(tz.dst(None), datetime.timedelta(0))

    def test_fromutc(self):
        import datetime

        naive_epoch = datetime.datetime(1970, 1, 1, 0, 0, 1, tzinfo=None)
        self.assertIsNone(naive_epoch.tzinfo)
        tz = self._make_one()
        epoch = tz.fromutc(naive_epoch)
        self.assertEqual(epoch.tzinfo, tz)

    def test_fromutc_with_tz(self):
        import datetime

        tz = self._make_one()
        epoch_with_tz = datetime.datetime(1970, 1, 1, 0, 0, 1, tzinfo=tz)
        epoch = tz.fromutc(epoch_with_tz)
        self.assertEqual(epoch.tzinfo, tz)

    def test_tzname(self):
        tz = self._make_one()
        self.assertEqual(tz.tzname(None), "UTC")

    def test_utcoffset(self):
        import datetime

        tz = self._make_one()
        self.assertEqual(tz.utcoffset(None), datetime.timedelta(0))

    def test___repr__(self):
        tz = self._make_one()
        self.assertEqual(repr(tz), "<UTC>")

    def test___str__(self):
        tz = self._make_one()
        self.assertEqual(str(tz), "UTC")


class Test__ensure_tuple_or_list(unittest.TestCase):
    def _call_fut(self, arg_name, tuple_or_list):
        from google.cloud._helpers import _ensure_tuple_or_list

        return _ensure_tuple_or_list(arg_name, tuple_or_list)

    def test_valid_tuple(self):
        valid_tuple_or_list = ("a", "b", "c", "d")
        result = self._call_fut("ARGNAME", valid_tuple_or_list)
        self.assertEqual(result, ["a", "b", "c", "d"])

    def test_valid_list(self):
        valid_tuple_or_list = ["a", "b", "c", "d"]
        result = self._call_fut("ARGNAME", valid_tuple_or_list)
        self.assertEqual(result, valid_tuple_or_list)

    def test_invalid(self):
        invalid_tuple_or_list = object()
        with self.assertRaises(TypeError):
            self._call_fut("ARGNAME", invalid_tuple_or_list)

    def test_invalid_iterable(self):
        invalid_tuple_or_list = "FOO"
        with self.assertRaises(TypeError):
            self._call_fut("ARGNAME", invalid_tuple_or_list)


class Test__determine_default_project(unittest.TestCase):
    def _call_fut(self, project=None):
        from google.cloud._helpers import _determine_default_project

        return _determine_default_project(project=project)

    def test_it(self):
        with mock.patch("google.auth.default", autospec=True) as default:
            default.return_value = (mock.sentinel.credentials, mock.sentinel.project)
            project = self._call_fut()

        self.assertEqual(project, mock.sentinel.project)
        default.assert_called_once_with()

    def test_explicit(self):
        with mock.patch("google.auth.default", autospec=True) as default:
            project = self._call_fut(mock.sentinel.project)

        self.assertEqual(project, mock.sentinel.project)
        self.assertFalse(default.called)


class Test__millis(unittest.TestCase):
    def _call_fut(self, value):
        from google.cloud._helpers import _millis

        return _millis(value)

    def test_one_second_from_epoch(self):
        import datetime
        from google.cloud._helpers import UTC

        WHEN = datetime.datetime(1970, 1, 1, 0, 0, 1, tzinfo=UTC)
        self.assertEqual(self._call_fut(WHEN), 1000)


class Test__microseconds_from_datetime(unittest.TestCase):
    def _call_fut(self, value):
        from google.cloud._helpers import _microseconds_from_datetime

        return _microseconds_from_datetime(value)

    def test_it(self):
        import datetime

        microseconds = 314159
        timestamp = datetime.datetime(
            1970, 1, 1, hour=0, minute=0, second=0, microsecond=microseconds
        )
        result = self._call_fut(timestamp)
        self.assertEqual(result, microseconds)


class Test__millis_from_datetime(unittest.TestCase):
    def _call_fut(self, value):
        from google.cloud._helpers import _millis_from_datetime

        return _millis_from_datetime(value)

    def test_w_none(self):
        self.assertIsNone(self._call_fut(None))

    def test_w_utc_datetime(self):
        import datetime
        import six
        from google.cloud._helpers import UTC
        from google.cloud._helpers import _microseconds_from_datetime

        NOW = datetime.datetime.utcnow().replace(tzinfo=UTC)
        NOW_MICROS = _microseconds_from_datetime(NOW)
        MILLIS = NOW_MICROS // 1000
        result = self._call_fut(NOW)
        self.assertIsInstance(result, six.integer_types)
        self.assertEqual(result, MILLIS)

    def test_w_non_utc_datetime(self):
        import datetime
        import six
        from google.cloud._helpers import _UTC
        from google.cloud._helpers import _microseconds_from_datetime

        class CET(_UTC):
            _tzname = "CET"
            _utcoffset = datetime.timedelta(hours=-1)

        zone = CET()
        NOW = datetime.datetime(2015, 7, 28, 16, 34, 47, tzinfo=zone)
        NOW_MICROS = _microseconds_from_datetime(NOW)
        MILLIS = NOW_MICROS // 1000
        result = self._call_fut(NOW)
        self.assertIsInstance(result, six.integer_types)
        self.assertEqual(result, MILLIS)

    def test_w_naive_datetime(self):
        import datetime
        import six
        from google.cloud._helpers import UTC
        from google.cloud._helpers import _microseconds_from_datetime

        NOW = datetime.datetime.utcnow()
        UTC_NOW = NOW.replace(tzinfo=UTC)
        UTC_NOW_MICROS = _microseconds_from_datetime(UTC_NOW)
        MILLIS = UTC_NOW_MICROS // 1000
        result = self._call_fut(NOW)
        self.assertIsInstance(result, six.integer_types)
        self.assertEqual(result, MILLIS)


class Test__datetime_from_microseconds(unittest.TestCase):
    def _call_fut(self, value):
        from google.cloud._helpers import _datetime_from_microseconds

        return _datetime_from_microseconds(value)

    def test_it(self):
        import datetime
        from google.cloud._helpers import UTC
        from google.cloud._helpers import _microseconds_from_datetime

        NOW = datetime.datetime(2015, 7, 29, 17, 45, 21, 123456, tzinfo=UTC)
        NOW_MICROS = _microseconds_from_datetime(NOW)
        self.assertEqual(self._call_fut(NOW_MICROS), NOW)


class Test___date_from_iso8601_date(unittest.TestCase):
    def _call_fut(self, value):
        from google.cloud._helpers import _date_from_iso8601_date

        return _date_from_iso8601_date(value)

    def test_todays_date(self):
        import datetime

        TODAY = datetime.date.today()
        self.assertEqual(self._call_fut(TODAY.strftime("%Y-%m-%d")), TODAY)


class Test___time_from_iso8601_time_naive(unittest.TestCase):
    def _call_fut(self, value):
        from google.cloud._helpers import _time_from_iso8601_time_naive

        return _time_from_iso8601_time_naive(value)

    def test_todays_date(self):
        import datetime

        WHEN = datetime.time(12, 9, 42)
        self.assertEqual(self._call_fut(("12:09:42")), WHEN)

    def test_w_microseconds(self):
        import datetime

        WHEN = datetime.time(12, 9, 42, 123456)
        self.assertEqual(self._call_fut(("12:09:42.123456")), WHEN)

    def test_w_millis_fail(self):
        with self.assertRaises(ValueError):
            self._call_fut("12:09:42.123")


class Test__rfc3339_to_datetime(unittest.TestCase):
    def _call_fut(self, dt_str):
        from google.cloud._helpers import _rfc3339_to_datetime

        return _rfc3339_to_datetime(dt_str)

    def test_w_bogus_zone(self):
        year = 2009
        month = 12
        day = 17
        hour = 12
        minute = 44
        seconds = 32
        micros = 123456789

        dt_str = "%d-%02d-%02dT%02d:%02d:%02d.%06dBOGUS" % (
            year,
            month,
            day,
            hour,
            minute,
            seconds,
            micros,
        )
        with self.assertRaises(ValueError):
            self._call_fut(dt_str)

    def test_w_microseconds(self):
        import datetime
        from google.cloud._helpers import UTC

        year = 2009
        month = 12
        day = 17
        hour = 12
        minute = 44
        seconds = 32
        micros = 123456

        dt_str = "%d-%02d-%02dT%02d:%02d:%02d.%06dZ" % (
            year,
            month,
            day,
            hour,
            minute,
            seconds,
            micros,
        )
        result = self._call_fut(dt_str)
        expected_result = datetime.datetime(
            year, month, day, hour, minute, seconds, micros, UTC
        )
        self.assertEqual(result, expected_result)

    def test_w_naonseconds(self):
        year = 2009
        month = 12
        day = 17
        hour = 12
        minute = 44
        seconds = 32
        nanos = 123456789

        dt_str = "%d-%02d-%02dT%02d:%02d:%02d.%09dZ" % (
            year,
            month,
            day,
            hour,
            minute,
            seconds,
            nanos,
        )
        with self.assertRaises(ValueError):
            self._call_fut(dt_str)


class Test__rfc3339_nanos_to_datetime(unittest.TestCase):
    def _call_fut(self, dt_str):
        from google.cloud._helpers import _rfc3339_nanos_to_datetime

        return _rfc3339_nanos_to_datetime(dt_str)

    def test_w_bogus_zone(self):
        year = 2009
        month = 12
        day = 17
        hour = 12
        minute = 44
        seconds = 32
        micros = 123456789

        dt_str = "%d-%02d-%02dT%02d:%02d:%02d.%06dBOGUS" % (
            year,
            month,
            day,
            hour,
            minute,
            seconds,
            micros,
        )
        with self.assertRaises(ValueError):
            self._call_fut(dt_str)

    def test_w_truncated_nanos(self):
        import datetime
        from google.cloud._helpers import UTC

        year = 2009
        month = 12
        day = 17
        hour = 12
        minute = 44
        seconds = 32
        truncateds_and_micros = [
            ("12345678", 123456),
            ("1234567", 123456),
            ("123456", 123456),
            ("12345", 123450),
            ("1234", 123400),
            ("123", 123000),
            ("12", 120000),
            ("1", 100000),
        ]

        for truncated, micros in truncateds_and_micros:
            dt_str = "%d-%02d-%02dT%02d:%02d:%02d.%sZ" % (
                year,
                month,
                day,
                hour,
                minute,
                seconds,
                truncated,
            )
            result = self._call_fut(dt_str)
            expected_result = datetime.datetime(
                year, month, day, hour, minute, seconds, micros, UTC
            )
            self.assertEqual(result, expected_result)

    def test_without_nanos(self):
        import datetime
        from google.cloud._helpers import UTC

        year = 1988
        month = 4
        day = 29
        hour = 12
        minute = 12
        seconds = 12

        dt_str = "%d-%02d-%02dT%02d:%02d:%02dZ" % (
            year,
            month,
            day,
            hour,
            minute,
            seconds,
        )
        result = self._call_fut(dt_str)
        expected_result = datetime.datetime(
            year, month, day, hour, minute, seconds, 0, UTC
        )
        self.assertEqual(result, expected_result)

    def test_w_naonseconds(self):
        import datetime
        from google.cloud._helpers import UTC

        year = 2009
        month = 12
        day = 17
        hour = 12
        minute = 44
        seconds = 32
        nanos = 123456789
        micros = nanos // 1000

        dt_str = "%d-%02d-%02dT%02d:%02d:%02d.%09dZ" % (
            year,
            month,
            day,
            hour,
            minute,
            seconds,
            nanos,
        )
        result = self._call_fut(dt_str)
        expected_result = datetime.datetime(
            year, month, day, hour, minute, seconds, micros, UTC
        )
        self.assertEqual(result, expected_result)


class Test__datetime_to_rfc3339(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import _datetime_to_rfc3339

        return _datetime_to_rfc3339(*args, **kwargs)

    @staticmethod
    def _make_timezone(offset):
        from google.cloud._helpers import _UTC

        class CET(_UTC):
            _tzname = "CET"
            _utcoffset = offset

        return CET()

    def test_w_utc_datetime(self):
        import datetime
        from google.cloud._helpers import UTC

        TIMESTAMP = datetime.datetime(2016, 4, 5, 13, 30, 0, tzinfo=UTC)
        result = self._call_fut(TIMESTAMP, ignore_zone=False)
        self.assertEqual(result, "2016-04-05T13:30:00.000000Z")

    def test_w_non_utc_datetime(self):
        import datetime

        zone = self._make_timezone(offset=datetime.timedelta(hours=-1))
        TIMESTAMP = datetime.datetime(2016, 4, 5, 13, 30, 0, tzinfo=zone)
        result = self._call_fut(TIMESTAMP, ignore_zone=False)
        self.assertEqual(result, "2016-04-05T14:30:00.000000Z")

    def test_w_non_utc_datetime_and_ignore_zone(self):
        import datetime

        zone = self._make_timezone(offset=datetime.timedelta(hours=-1))
        TIMESTAMP = datetime.datetime(2016, 4, 5, 13, 30, 0, tzinfo=zone)
        result = self._call_fut(TIMESTAMP)
        self.assertEqual(result, "2016-04-05T13:30:00.000000Z")

    def test_w_naive_datetime(self):
        import datetime

        TIMESTAMP = datetime.datetime(2016, 4, 5, 13, 30, 0)
        result = self._call_fut(TIMESTAMP)
        self.assertEqual(result, "2016-04-05T13:30:00.000000Z")


class Test__to_bytes(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import _to_bytes

        return _to_bytes(*args, **kwargs)

    def test_with_bytes(self):
        value = b"bytes-val"
        self.assertEqual(self._call_fut(value), value)

    def test_with_unicode(self):
        value = u"string-val"
        encoded_value = b"string-val"
        self.assertEqual(self._call_fut(value), encoded_value)

    def test_unicode_non_ascii(self):
        value = u"\u2013"  # Long hyphen
        encoded_value = b"\xe2\x80\x93"
        self.assertRaises(UnicodeEncodeError, self._call_fut, value)
        self.assertEqual(self._call_fut(value, encoding="utf-8"), encoded_value)

    def test_with_nonstring_type(self):
        value = object()
        self.assertRaises(TypeError, self._call_fut, value)


class Test__bytes_to_unicode(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import _bytes_to_unicode

        return _bytes_to_unicode(*args, **kwargs)

    def test_with_bytes(self):
        value = b"bytes-val"
        encoded_value = "bytes-val"
        self.assertEqual(self._call_fut(value), encoded_value)

    def test_with_unicode(self):
        value = u"string-val"
        encoded_value = "string-val"
        self.assertEqual(self._call_fut(value), encoded_value)

    def test_with_nonstring_type(self):
        value = object()
        self.assertRaises(ValueError, self._call_fut, value)


class Test__pb_timestamp_to_datetime(unittest.TestCase):
    def _call_fut(self, timestamp):
        from google.cloud._helpers import _pb_timestamp_to_datetime

        return _pb_timestamp_to_datetime(timestamp)

    def test_it(self):
        import datetime
        from google.protobuf.timestamp_pb2 import Timestamp
        from google.cloud._helpers import UTC

        # Epoch is midnight on January 1, 1970 ...
        dt_stamp = datetime.datetime(
            1970,
            month=1,
            day=1,
            hour=0,
            minute=1,
            second=1,
            microsecond=1234,
            tzinfo=UTC,
        )
        # ... so 1 minute and 1 second after is 61 seconds and 1234
        # microseconds is 1234000 nanoseconds.
        timestamp = Timestamp(seconds=61, nanos=1234000)
        self.assertEqual(self._call_fut(timestamp), dt_stamp)


class Test__from_any_pb(unittest.TestCase):
    def _call_fut(self, pb_type, any_pb):
        from google.cloud._helpers import _from_any_pb

        return _from_any_pb(pb_type, any_pb)

    def test_success(self):
        from google.protobuf import any_pb2
        from google.type import date_pb2

        in_message = date_pb2.Date(year=1990)
        in_message_any = any_pb2.Any()
        in_message_any.Pack(in_message)
        out_message = self._call_fut(date_pb2.Date, in_message_any)
        self.assertEqual(in_message, out_message)

    def test_failure(self,):
        from google.protobuf import any_pb2
        from google.type import date_pb2
        from google.type import timeofday_pb2

        in_message = any_pb2.Any()
        in_message.Pack(date_pb2.Date(year=1990))

        with self.assertRaises(TypeError):
            self._call_fut(timeofday_pb2.TimeOfDay, in_message)


class Test__pb_timestamp_to_rfc3339(unittest.TestCase):
    def _call_fut(self, timestamp):
        from google.cloud._helpers import _pb_timestamp_to_rfc3339

        return _pb_timestamp_to_rfc3339(timestamp)

    def test_it(self):
        from google.protobuf.timestamp_pb2 import Timestamp

        # Epoch is midnight on January 1, 1970 ...
        # ... so 1 minute and 1 second after is 61 seconds and 1234
        # microseconds is 1234000 nanoseconds.
        timestamp = Timestamp(seconds=61, nanos=1234000)
        self.assertEqual(self._call_fut(timestamp), "1970-01-01T00:01:01.001234Z")


class Test__datetime_to_pb_timestamp(unittest.TestCase):
    def _call_fut(self, when):
        from google.cloud._helpers import _datetime_to_pb_timestamp

        return _datetime_to_pb_timestamp(when)

    def test_it(self):
        import datetime
        from google.protobuf.timestamp_pb2 import Timestamp
        from google.cloud._helpers import UTC

        # Epoch is midnight on January 1, 1970 ...
        dt_stamp = datetime.datetime(
            1970,
            month=1,
            day=1,
            hour=0,
            minute=1,
            second=1,
            microsecond=1234,
            tzinfo=UTC,
        )
        # ... so 1 minute and 1 second after is 61 seconds and 1234
        # microseconds is 1234000 nanoseconds.
        timestamp = Timestamp(seconds=61, nanos=1234000)
        self.assertEqual(self._call_fut(dt_stamp), timestamp)


class Test__timedelta_to_duration_pb(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import _timedelta_to_duration_pb

        return _timedelta_to_duration_pb(*args, **kwargs)

    def test_it(self):
        import datetime
        from google.protobuf import duration_pb2

        seconds = microseconds = 1
        timedelta_val = datetime.timedelta(seconds=seconds, microseconds=microseconds)
        result = self._call_fut(timedelta_val)
        self.assertIsInstance(result, duration_pb2.Duration)
        self.assertEqual(result.seconds, seconds)
        self.assertEqual(result.nanos, 1000 * microseconds)

    def test_with_negative_microseconds(self):
        import datetime
        from google.protobuf import duration_pb2

        seconds = 1
        microseconds = -5
        timedelta_val = datetime.timedelta(seconds=seconds, microseconds=microseconds)
        result = self._call_fut(timedelta_val)
        self.assertIsInstance(result, duration_pb2.Duration)
        self.assertEqual(result.seconds, seconds - 1)
        self.assertEqual(result.nanos, 10 ** 9 + 1000 * microseconds)

    def test_with_negative_seconds(self):
        import datetime
        from google.protobuf import duration_pb2

        seconds = -1
        microseconds = 5
        timedelta_val = datetime.timedelta(seconds=seconds, microseconds=microseconds)
        result = self._call_fut(timedelta_val)
        self.assertIsInstance(result, duration_pb2.Duration)
        self.assertEqual(result.seconds, seconds + 1)
        self.assertEqual(result.nanos, -(10 ** 9 - 1000 * microseconds))


class Test__duration_pb_to_timedelta(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import _duration_pb_to_timedelta

        return _duration_pb_to_timedelta(*args, **kwargs)

    def test_it(self):
        import datetime
        from google.protobuf import duration_pb2

        seconds = microseconds = 1
        duration_pb = duration_pb2.Duration(seconds=seconds, nanos=1000 * microseconds)
        timedelta_val = datetime.timedelta(seconds=seconds, microseconds=microseconds)
        result = self._call_fut(duration_pb)
        self.assertIsInstance(result, datetime.timedelta)
        self.assertEqual(result, timedelta_val)


class Test__name_from_project_path(unittest.TestCase):

    PROJECT = "PROJECT"
    THING_NAME = "THING_NAME"
    TEMPLATE = r"projects/(?P<project>\w+)/things/(?P<name>\w+)"

    def _call_fut(self, path, project, template):
        from google.cloud._helpers import _name_from_project_path

        return _name_from_project_path(path, project, template)

    def test_w_invalid_path_length(self):
        PATH = "projects/foo"
        with self.assertRaises(ValueError):
            self._call_fut(PATH, None, self.TEMPLATE)

    def test_w_invalid_path_segments(self):
        PATH = "foo/%s/bar/%s" % (self.PROJECT, self.THING_NAME)
        with self.assertRaises(ValueError):
            self._call_fut(PATH, self.PROJECT, self.TEMPLATE)

    def test_w_mismatched_project(self):
        PROJECT1 = "PROJECT1"
        PROJECT2 = "PROJECT2"
        PATH = "projects/%s/things/%s" % (PROJECT1, self.THING_NAME)
        with self.assertRaises(ValueError):
            self._call_fut(PATH, PROJECT2, self.TEMPLATE)

    def test_w_valid_data_w_compiled_regex(self):
        import re

        template = re.compile(self.TEMPLATE)
        PATH = "projects/%s/things/%s" % (self.PROJECT, self.THING_NAME)
        name = self._call_fut(PATH, self.PROJECT, template)
        self.assertEqual(name, self.THING_NAME)

    def test_w_project_passed_as_none(self):
        PROJECT1 = "PROJECT1"
        PATH = "projects/%s/things/%s" % (PROJECT1, self.THING_NAME)
        self._call_fut(PATH, None, self.TEMPLATE)
        name = self._call_fut(PATH, None, self.TEMPLATE)
        self.assertEqual(name, self.THING_NAME)


class Test_make_secure_channel(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import make_secure_channel

        return make_secure_channel(*args, **kwargs)

    def test_it(self):
        from six.moves import http_client

        credentials = object()
        host = "HOST"
        user_agent = "USER_AGENT"

        secure_authorized_channel_patch = mock.patch(
            "google.auth.transport.grpc.secure_authorized_channel", autospec=True
        )

        with secure_authorized_channel_patch as secure_authorized_channel:
            result = self._call_fut(credentials, user_agent, host)

        self.assertIs(result, secure_authorized_channel.return_value)

        expected_target = "%s:%d" % (host, http_client.HTTPS_PORT)
        expected_options = (("grpc.primary_user_agent", user_agent),)

        secure_authorized_channel.assert_called_once_with(
            credentials, mock.ANY, expected_target, options=expected_options
        )

    def test_extra_options(self):
        from six.moves import http_client

        credentials = object()
        host = "HOST"
        user_agent = "USER_AGENT"
        extra_options = (("some", "option"),)

        secure_authorized_channel_patch = mock.patch(
            "google.auth.transport.grpc.secure_authorized_channel", autospec=True
        )

        with secure_authorized_channel_patch as secure_authorized_channel:
            result = self._call_fut(credentials, user_agent, host, extra_options)

        self.assertIs(result, secure_authorized_channel.return_value)

        expected_target = "%s:%d" % (host, http_client.HTTPS_PORT)
        expected_options = (("grpc.primary_user_agent", user_agent), extra_options[0])

        secure_authorized_channel.assert_called_once_with(
            credentials, mock.ANY, expected_target, options=expected_options
        )


class Test_make_secure_stub(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import make_secure_stub

        return make_secure_stub(*args, **kwargs)

    def test_it(self):
        from google.cloud._testing import _Monkey
        from google.cloud import _helpers as MUT

        result = object()
        channel_obj = object()
        channels = []
        channel_args = []

        def stub_class(channel):
            channels.append(channel)
            return result

        def mock_channel(*args, **kwargs):
            channel_args.append(args)
            channel_args.append(kwargs)
            return channel_obj

        credentials = object()
        user_agent = "you-sir-age-int"
        host = "localhost"
        extra_options = {"extra_options": ()}
        with _Monkey(MUT, make_secure_channel=mock_channel):
            stub = self._call_fut(credentials, user_agent, stub_class, host)

        self.assertIs(stub, result)
        self.assertEqual(channels, [channel_obj])
        self.assertEqual(channel_args, [(credentials, user_agent, host), extra_options])


class Test_make_insecure_stub(unittest.TestCase):
    def _call_fut(self, *args, **kwargs):
        from google.cloud._helpers import make_insecure_stub

        return make_insecure_stub(*args, **kwargs)

    def _helper(self, target, host, port=None):
        from google.cloud._testing import _Monkey
        from google.cloud import _helpers as MUT

        mock_result = object()
        stub_inputs = []
        CHANNEL = object()

        class _GRPCModule(object):
            def insecure_channel(self, *args):
                self.insecure_channel_args = args
                return CHANNEL

        grpc_mod = _GRPCModule()

        def mock_stub_class(channel):
            stub_inputs.append(channel)
            return mock_result

        with _Monkey(MUT, grpc=grpc_mod):
            result = self._call_fut(mock_stub_class, host, port=port)

        self.assertIs(result, mock_result)
        self.assertEqual(stub_inputs, [CHANNEL])
        self.assertEqual(grpc_mod.insecure_channel_args, (target,))

    def test_with_port_argument(self):
        host = "HOST"
        port = 1025
        target = "%s:%d" % (host, port)
        self._helper(target, host, port=port)

    def test_without_port_argument(self):
        host = "HOST:1114"
        self._helper(host, host)



# Copyright 2014 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import unittest
import warnings

import mock
import requests
from six.moves import http_client


class TestConnection(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from google.cloud._http import Connection

        return Connection

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def test_constructor_defaults(self):
        from google.api_core.client_info import ClientInfo

        client = object()
        conn = self._make_one(client)
        self.assertIs(conn._client, client)
        self.assertIsInstance(conn._client_info, ClientInfo)

    def test_constructor_explicit(self):
        client = object()
        client_info = object()
        conn = self._make_one(client, client_info=client_info)
        self.assertIs(conn._client, client)

    def test_user_agent_all_caps_getter_deprecated(self):
        client = object()
        conn = self._make_one(client)

        with mock.patch.object(warnings, "warn", autospec=True) as warn:
            self.assertEqual(conn.USER_AGENT, conn._client_info.to_user_agent())

        warn.assert_called_once_with(mock.ANY, DeprecationWarning, stacklevel=2)

    def test_user_agent_all_caps_setter_deprecated(self):
        conn = self._make_one(object())
        user_agent = "testing"

        with mock.patch.object(warnings, "warn", autospec=True) as warn:
            conn.USER_AGENT = user_agent

        self.assertEqual(conn._client_info.user_agent, user_agent)
        warn.assert_called_once_with(mock.ANY, DeprecationWarning, stacklevel=2)

    def test_user_agent_getter(self):
        conn = self._make_one(object())
        self.assertEqual(conn.user_agent, conn._client_info.to_user_agent())

    def test_user_agent_setter(self):
        conn = self._make_one(object())
        user_agent = "testing"
        conn.user_agent = user_agent
        self.assertEqual(conn._client_info.user_agent, user_agent)

    def test_extra_headers_all_caps_getter_deprecated(self):
        client = object()
        conn = self._make_one(client)
        expected = conn._extra_headers = {"foo": "bar"}

        with mock.patch.object(warnings, "warn", autospec=True) as warn:
            self.assertEqual(conn._EXTRA_HEADERS, expected)

        warn.assert_called_once_with(mock.ANY, DeprecationWarning, stacklevel=2)

    def test_extra_headers_all_caps_setter_deprecated(self):
        conn = self._make_one(object())
        extra_headers = {"foo": "bar"}

        with mock.patch.object(warnings, "warn", autospec=True) as warn:
            conn._EXTRA_HEADERS = extra_headers

        self.assertEqual(conn._extra_headers, extra_headers)
        warn.assert_called_once_with(mock.ANY, DeprecationWarning, stacklevel=2)

    def test_extra_headers_getter_default(self):
        conn = self._make_one(object())
        expected = {}
        self.assertEqual(conn.extra_headers, expected)

    def test_extra_headers_getter_overridden(self):
        conn = self._make_one(object())
        expected = conn._extra_headers = {"foo": "bar"}
        self.assertEqual(conn.extra_headers, expected)

    def test_extra_headers_item_assignment(self):
        conn = self._make_one(object())
        expected = {"foo": "bar"}
        conn.extra_headers["foo"] = "bar"
        self.assertEqual(conn._extra_headers, expected)

    def test_extra_headers_setter(self):
        conn = self._make_one(object())
        expected = {"foo": "bar"}
        conn.extra_headers = expected
        self.assertEqual(conn._extra_headers, expected)

    def test_credentials_property(self):
        client = mock.Mock(spec=["_credentials"])
        conn = self._make_one(client)
        self.assertIs(conn.credentials, client._credentials)

    def test_http_property(self):
        client = mock.Mock(spec=["_http"])
        conn = self._make_one(client)
        self.assertIs(conn.http, client._http)


def make_response(status=http_client.OK, content=b"", headers={}):
    response = requests.Response()
    response.status_code = status
    response._content = content
    response.headers = headers
    response.request = requests.Request()
    return response


def make_requests_session(responses):
    session = mock.create_autospec(requests.Session, instance=True)
    session.request.side_effect = responses
    return session


class TestJSONConnection(unittest.TestCase):
    JSON_HEADERS = {"content-type": "application/json"}
    EMPTY_JSON_RESPONSE = make_response(content=b"{}", headers=JSON_HEADERS)

    @staticmethod
    def _get_default_timeout():
        from google.cloud._http import _DEFAULT_TIMEOUT

        return _DEFAULT_TIMEOUT

    @staticmethod
    def _get_target_class():
        from google.cloud._http import JSONConnection

        return JSONConnection

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def _make_mock_one(self, *args, **kw):
        class MockConnection(self._get_target_class()):
            API_URL_TEMPLATE = "{api_base_url}/mock/{api_version}{path}"
            API_BASE_URL = "http://mock"
            API_BASE_MTLS_URL = "https://mock.mtls"
            API_VERSION = "vMOCK"

        return MockConnection(*args, **kw)

    def test_class_defaults(self):
        klass = self._get_target_class()
        self.assertIsNone(klass.API_URL_TEMPLATE)
        self.assertIsNone(klass.API_BASE_URL)
        self.assertIsNone(klass.API_VERSION)

    def test_constructor(self):
        client = object()
        conn = self._make_one(client)
        self.assertIs(conn._client, client)

    def test_build_api_url_no_extra_query_params(self):
        client = object()
        conn = self._make_mock_one(client)
        # Intended to emulate self.mock_template
        URI = "/".join(
            [conn.API_BASE_URL, "mock", conn.API_VERSION, "foo?prettyPrint=false"]
        )
        self.assertEqual(conn.build_api_url("/foo"), URI)

    def test_build_api_url_w_pretty_print_query_params(self):
        client = object()
        conn = self._make_mock_one(client)
        uri = conn.build_api_url("/foo", {"prettyPrint": "true"})
        URI = "/".join(
            [conn.API_BASE_URL, "mock", conn.API_VERSION, "foo?prettyPrint=true"]
        )
        self.assertEqual(uri, URI)

    def test_build_api_url_w_extra_query_params(self):
        from six.moves.urllib.parse import parse_qs
        from six.moves.urllib.parse import urlsplit

        client = object()
        conn = self._make_mock_one(client)
        uri = conn.build_api_url("/foo", {"bar": "baz", "qux": ["quux", "corge"]})

        scheme, netloc, path, qs, _ = urlsplit(uri)
        self.assertEqual("%s://%s" % (scheme, netloc), conn.API_BASE_URL)
        # Intended to emulate mock_template
        PATH = "/".join(["", "mock", conn.API_VERSION, "foo"])
        self.assertEqual(path, PATH)
        parms = dict(parse_qs(qs))
        self.assertEqual(parms["bar"], ["baz"])
        self.assertEqual(parms["qux"], ["quux", "corge"])
        self.assertEqual(parms["prettyPrint"], ["false"])

    def test_build_api_url_w_extra_query_params_tuples(self):
        from six.moves.urllib.parse import parse_qs
        from six.moves.urllib.parse import urlsplit

        client = object()
        conn = self._make_mock_one(client)
        uri = conn.build_api_url(
            "/foo", [("bar", "baz"), ("qux", "quux"), ("qux", "corge")]
        )

        scheme, netloc, path, qs, _ = urlsplit(uri)
        self.assertEqual("%s://%s" % (scheme, netloc), conn.API_BASE_URL)
        # Intended to emulate mock_template
        PATH = "/".join(["", "mock", conn.API_VERSION, "foo"])
        self.assertEqual(path, PATH)
        parms = dict(parse_qs(qs))
        self.assertEqual(parms["bar"], ["baz"])
        self.assertEqual(parms["qux"], ["quux", "corge"])
        self.assertEqual(parms["prettyPrint"], ["false"])

    def test_get_api_base_url_for_mtls_w_api_base_url(self):
        client = object()
        conn = self._make_mock_one(client)
        uri = conn.get_api_base_url_for_mtls(api_base_url="http://foo")
        self.assertEqual(uri, "http://foo")

    def test_get_api_base_url_for_mtls_env_always(self):
        client = object()
        conn = self._make_mock_one(client)
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "always"}):
            uri = conn.get_api_base_url_for_mtls()
            self.assertEqual(uri, "https://mock.mtls")

    def test_get_api_base_url_for_mtls_env_never(self):
        client = object()
        conn = self._make_mock_one(client)
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "never"}):
            uri = conn.get_api_base_url_for_mtls()
            self.assertEqual(uri, "http://mock")

    def test_get_api_base_url_for_mtls_env_auto(self):
        client = mock.Mock()
        client._http = mock.Mock()
        client._http.is_mtls = False
        conn = self._make_mock_one(client)

        # ALLOW_AUTO_SWITCH_TO_MTLS_URL is False, so use regular endpoint.
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "auto"}):
            uri = conn.get_api_base_url_for_mtls()
            self.assertEqual(uri, "http://mock")

        # ALLOW_AUTO_SWITCH_TO_MTLS_URL is True, so now endpoint dependes
        # on client._http.is_mtls
        conn.ALLOW_AUTO_SWITCH_TO_MTLS_URL = True

        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "auto"}):
            uri = conn.get_api_base_url_for_mtls()
            self.assertEqual(uri, "http://mock")

        client._http.is_mtls = True
        with mock.patch.dict(os.environ, {"GOOGLE_API_USE_MTLS_ENDPOINT": "auto"}):
            uri = conn.get_api_base_url_for_mtls()
            self.assertEqual(uri, "https://mock.mtls")

    def test__make_request_no_data_no_content_type_no_headers(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([make_response()])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_one(client)
        url = "http://example.com/test"

        response = conn._make_request("GET", url)

        self.assertEqual(response.status_code, http_client.OK)
        self.assertEqual(response.content, b"")

        expected_headers = {
            "Accept-Encoding": "gzip",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        http.request.assert_called_once_with(
            method="GET",
            url=url,
            headers=expected_headers,
            data=None,
            timeout=self._get_default_timeout(),
        )

    def test__make_request_w_data_no_extra_headers(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([make_response()])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_one(client)
        url = "http://example.com/test"
        data = b"data"

        conn._make_request("GET", url, data, "application/json")

        expected_headers = {
            "Accept-Encoding": "gzip",
            "Content-Type": "application/json",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        http.request.assert_called_once_with(
            method="GET",
            url=url,
            headers=expected_headers,
            data=data,
            timeout=self._get_default_timeout(),
        )

    def test__make_request_w_extra_headers(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([make_response()])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_one(client)

        url = "http://example.com/test"
        conn._make_request("GET", url, headers={"X-Foo": "foo"})

        expected_headers = {
            "Accept-Encoding": "gzip",
            "X-Foo": "foo",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        http.request.assert_called_once_with(
            method="GET",
            url=url,
            headers=expected_headers,
            data=None,
            timeout=self._get_default_timeout(),
        )

    def test__make_request_w_timeout(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([make_response()])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_one(client)

        url = "http://example.com/test"
        conn._make_request("GET", url, timeout=(5.5, 2.8))

        expected_headers = {
            "Accept-Encoding": "gzip",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        http.request.assert_called_once_with(
            method="GET",
            url=url,
            headers=expected_headers,
            data=None,
            timeout=(5.5, 2.8),
        )

    def test_api_request_defaults(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session(
            [make_response(content=b"{}", headers=self.JSON_HEADERS)]
        )
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)
        path = "/path/required"

        self.assertEqual(conn.api_request("GET", path), {})

        expected_headers = {
            "Accept-Encoding": "gzip",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        expected_url = "{base}/mock/{version}{path}?prettyPrint=false".format(
            base=conn.API_BASE_URL, version=conn.API_VERSION, path=path
        )
        http.request.assert_called_once_with(
            method="GET",
            url=expected_url,
            headers=expected_headers,
            data=None,
            timeout=self._get_default_timeout(),
        )

    def test_api_request_w_non_json_response(self):
        http = make_requests_session([make_response(content=b"content")])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)

        with self.assertRaises(ValueError):
            conn.api_request("GET", "/")

    def test_api_request_wo_json_expected(self):
        http = make_requests_session([make_response(content=b"content")])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)

        result = conn.api_request("GET", "/", expect_json=False)

        self.assertEqual(result, b"content")

    def test_api_request_w_query_params(self):
        from six.moves.urllib.parse import parse_qs
        from six.moves.urllib.parse import urlsplit
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([self.EMPTY_JSON_RESPONSE])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)

        result = conn.api_request("GET", "/", {"foo": "bar", "baz": ["qux", "quux"]})

        self.assertEqual(result, {})

        expected_headers = {
            "Accept-Encoding": "gzip",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        http.request.assert_called_once_with(
            method="GET",
            url=mock.ANY,
            headers=expected_headers,
            data=None,
            timeout=self._get_default_timeout(),
        )

        url = http.request.call_args[1]["url"]
        scheme, netloc, path, qs, _ = urlsplit(url)
        self.assertEqual("%s://%s" % (scheme, netloc), conn.API_BASE_URL)
        # Intended to emulate self.mock_template
        PATH = "/".join(["", "mock", conn.API_VERSION, ""])
        self.assertEqual(path, PATH)
        parms = dict(parse_qs(qs))
        self.assertEqual(parms["foo"], ["bar"])
        self.assertEqual(parms["baz"], ["qux", "quux"])

    def test_api_request_w_headers(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([self.EMPTY_JSON_RESPONSE])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)

        result = conn.api_request("GET", "/", headers={"X-Foo": "bar"})
        self.assertEqual(result, {})

        expected_headers = {
            "Accept-Encoding": "gzip",
            "User-Agent": conn.user_agent,
            "X-Foo": "bar",
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        http.request.assert_called_once_with(
            method="GET",
            url=mock.ANY,
            headers=expected_headers,
            data=None,
            timeout=self._get_default_timeout(),
        )

    def test_api_request_w_extra_headers(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([self.EMPTY_JSON_RESPONSE])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)
        conn.extra_headers = {
            "X-Baz": "dax-quux",
            "X-Foo": "not-bar",  # Collision with ``headers``.
        }

        result = conn.api_request("GET", "/", headers={"X-Foo": "bar"})

        self.assertEqual(result, {})

        expected_headers = {
            "Accept-Encoding": "gzip",
            "User-Agent": conn.user_agent,
            "X-Foo": "not-bar",  # The one passed-in is overridden.
            "X-Baz": "dax-quux",
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        http.request.assert_called_once_with(
            method="GET",
            url=mock.ANY,
            headers=expected_headers,
            data=None,
            timeout=self._get_default_timeout(),
        )

    def test_api_request_w_data(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session([self.EMPTY_JSON_RESPONSE])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)

        data = {"foo": "bar"}
        self.assertEqual(conn.api_request("POST", "/", data=data), {})

        expected_data = json.dumps(data)

        expected_headers = {
            "Accept-Encoding": "gzip",
            "Content-Type": "application/json",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }

        http.request.assert_called_once_with(
            method="POST",
            url=mock.ANY,
            headers=expected_headers,
            data=expected_data,
            timeout=self._get_default_timeout(),
        )

    def test_api_request_w_timeout(self):
        from google.cloud._http import CLIENT_INFO_HEADER

        http = make_requests_session(
            [make_response(content=b"{}", headers=self.JSON_HEADERS)]
        )
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)
        path = "/path/required"

        self.assertEqual(conn.api_request("GET", path, timeout=(2.2, 3.3)), {})

        expected_headers = {
            "Accept-Encoding": "gzip",
            "User-Agent": conn.user_agent,
            CLIENT_INFO_HEADER: conn.user_agent,
        }
        expected_url = "{base}/mock/{version}{path}?prettyPrint=false".format(
            base=conn.API_BASE_URL, version=conn.API_VERSION, path=path
        )
        http.request.assert_called_once_with(
            method="GET",
            url=expected_url,
            headers=expected_headers,
            data=None,
            timeout=(2.2, 3.3),
        )

    def test_api_request_w_404(self):
        from google.cloud import exceptions

        http = make_requests_session([make_response(http_client.NOT_FOUND)])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)

        with self.assertRaises(exceptions.NotFound):
            conn.api_request("GET", "/")

    def test_api_request_w_500(self):
        from google.cloud import exceptions

        http = make_requests_session([make_response(http_client.INTERNAL_SERVER_ERROR)])
        client = mock.Mock(_http=http, spec=["_http"])
        conn = self._make_mock_one(client)

        with self.assertRaises(exceptions.InternalServerError):
            conn.api_request("GET", "/")



# Copyright 2016 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest


class Test__compute_type_url(unittest.TestCase):
    def _call_fut(self, klass, prefix=None):
        from google.cloud.operation import _compute_type_url

        if prefix is None:
            return _compute_type_url(klass)
        return _compute_type_url(klass, prefix)

    def test_wo_prefix(self):
        from google.protobuf.struct_pb2 import Struct
        from google.cloud.operation import _GOOGLE_APIS_PREFIX

        type_url = self._call_fut(Struct)

        self.assertEqual(
            type_url, "%s/%s" % (_GOOGLE_APIS_PREFIX, Struct.DESCRIPTOR.full_name)
        )

    def test_w_prefix(self):
        from google.protobuf.struct_pb2 import Struct

        PREFIX = "test.google-cloud-python.com"

        type_url = self._call_fut(Struct, PREFIX)

        self.assertEqual(type_url, "%s/%s" % (PREFIX, Struct.DESCRIPTOR.full_name))


class Test_register_type(unittest.TestCase):
    def _call_fut(self, klass, type_url=None):
        from google.cloud.operation import register_type

        register_type(klass, type_url=type_url)

    def test_explicit(self):
        from google.cloud import operation as MUT
        from google.cloud._testing import _Monkey

        type_url = "testing.google-cloud-python.com/testing"
        klass = object()
        type_url_map = {}

        with _Monkey(MUT, _TYPE_URL_MAP=type_url_map):
            self._call_fut(klass, type_url)

        self.assertEqual(type_url_map, {type_url: klass})

    def test_default(self):
        from google.protobuf.struct_pb2 import Struct
        from google.cloud._testing import _Monkey
        from google.cloud import operation as MUT

        type_url_map = {}
        with _Monkey(MUT, _TYPE_URL_MAP=type_url_map):
            self._call_fut(Struct)

        type_url = MUT._compute_type_url(Struct)
        self.assertEqual(type_url_map, {type_url: Struct})

    def test_w_same_class(self):
        from google.cloud import operation as MUT
        from google.cloud._testing import _Monkey

        type_url = "testing.google-cloud-python.com/testing"
        klass = object()
        type_url_map = {type_url: klass}

        with _Monkey(MUT, _TYPE_URL_MAP=type_url_map):
            self._call_fut(klass, type_url)

        self.assertEqual(type_url_map, {type_url: klass})

    def test_w_conflict(self):
        from google.cloud import operation as MUT
        from google.cloud._testing import _Monkey

        type_url = "testing.google-cloud-python.com/testing"
        klass, other = object(), object()
        type_url_map = {type_url: other}

        with _Monkey(MUT, _TYPE_URL_MAP=type_url_map):
            with self.assertRaises(ValueError):
                self._call_fut(klass, type_url)

        self.assertEqual(type_url_map, {type_url: other})


class TestOperation(unittest.TestCase):

    OPERATION_NAME = "operations/projects/foo/instances/bar/operations/123"

    @staticmethod
    def _get_target_class():
        from google.cloud.operation import Operation

        return Operation

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def test_ctor_defaults(self):
        client = _Client()
        operation = self._make_one(self.OPERATION_NAME, client)
        self.assertEqual(operation.name, self.OPERATION_NAME)
        self.assertIs(operation.client, client)
        self.assertIsNone(operation.target)
        self.assertIsNone(operation.response)
        self.assertIsNone(operation.error)
        self.assertIsNone(operation.metadata)
        self.assertEqual(operation.caller_metadata, {})
        self.assertTrue(operation._from_grpc)

    def test_ctor_explicit(self):
        client = _Client()
        operation = self._make_one(self.OPERATION_NAME, client, foo="bar")

        self.assertEqual(operation.name, self.OPERATION_NAME)
        self.assertIs(operation.client, client)
        self.assertIsNone(operation.target)
        self.assertIsNone(operation.response)
        self.assertIsNone(operation.error)
        self.assertIsNone(operation.metadata)
        self.assertEqual(operation.caller_metadata, {"foo": "bar"})
        self.assertTrue(operation._from_grpc)

    def test_from_pb_wo_metadata_or_kw(self):
        from google.longrunning import operations_pb2

        client = _Client()
        operation_pb = operations_pb2.Operation(name=self.OPERATION_NAME)
        klass = self._get_target_class()

        operation = klass.from_pb(operation_pb, client)

        self.assertEqual(operation.name, self.OPERATION_NAME)
        self.assertIs(operation.client, client)
        self.assertIsNone(operation.metadata)
        self.assertEqual(operation.caller_metadata, {})

    def test_from_pb_w_unknown_metadata(self):
        from google.longrunning import operations_pb2
        from google.protobuf.any_pb2 import Any
        from google.protobuf.json_format import ParseDict
        from google.protobuf.struct_pb2 import Struct
        from google.cloud._testing import _Monkey
        from google.cloud import operation as MUT

        type_url = "type.googleapis.com/%s" % (Struct.DESCRIPTOR.full_name,)
        client = _Client()
        meta = ParseDict({"foo": "Bar"}, Struct())
        metadata_pb = Any(type_url=type_url, value=meta.SerializeToString())
        operation_pb = operations_pb2.Operation(
            name=self.OPERATION_NAME, metadata=metadata_pb
        )
        klass = self._get_target_class()

        with _Monkey(MUT, _TYPE_URL_MAP={type_url: Struct}):
            operation = klass.from_pb(operation_pb, client)

        self.assertEqual(operation.name, self.OPERATION_NAME)
        self.assertIs(operation.client, client)
        self.assertEqual(operation.metadata, meta)
        self.assertEqual(operation.caller_metadata, {})

    def test_from_pb_w_metadata_and_kwargs(self):
        from google.longrunning import operations_pb2
        from google.protobuf.any_pb2 import Any
        from google.protobuf.struct_pb2 import Struct
        from google.protobuf.struct_pb2 import Value
        from google.cloud import operation as MUT
        from google.cloud._testing import _Monkey

        type_url = "type.googleapis.com/%s" % (Struct.DESCRIPTOR.full_name,)
        type_url_map = {type_url: Struct}

        client = _Client()
        meta = Struct(fields={"foo": Value(string_value=u"Bar")})
        metadata_pb = Any(type_url=type_url, value=meta.SerializeToString())
        operation_pb = operations_pb2.Operation(
            name=self.OPERATION_NAME, metadata=metadata_pb
        )
        klass = self._get_target_class()

        with _Monkey(MUT, _TYPE_URL_MAP=type_url_map):
            operation = klass.from_pb(operation_pb, client, baz="qux")

        self.assertEqual(operation.name, self.OPERATION_NAME)
        self.assertIs(operation.client, client)
        self.assertEqual(operation.metadata, meta)
        self.assertEqual(operation.caller_metadata, {"baz": "qux"})

    def test_from_dict(self):
        from google.protobuf.struct_pb2 import Struct
        from google.cloud._testing import _Monkey
        from google.cloud import operation as MUT

        type_url = "type.googleapis.com/%s" % (Struct.DESCRIPTOR.full_name,)
        api_response = {
            "name": self.OPERATION_NAME,
            "metadata": {"@type": type_url, "value": {"foo": "Bar"}},
        }

        client = _Client()
        klass = self._get_target_class()

        with _Monkey(MUT, _TYPE_URL_MAP={type_url: Struct}):
            operation = klass.from_dict(api_response, client)

        self.assertEqual(operation.name, self.OPERATION_NAME)
        self.assertIs(operation.client, client)
        self.assertIsNone(operation.target)
        self.assertIsNone(operation.response)
        self.assertIsNone(operation.error)
        self.assertIsInstance(operation.metadata, Struct)
        self.assertEqual(len(operation.metadata.fields), 1)
        self.assertEqual(operation.metadata.fields["foo"].string_value, "Bar")
        self.assertEqual(operation.caller_metadata, {})
        self.assertFalse(operation._from_grpc)

    def test_complete_property(self):
        client = _Client()
        operation = self._make_one(self.OPERATION_NAME, client)
        self.assertFalse(operation.complete)
        operation._complete = True
        self.assertTrue(operation.complete)
        with self.assertRaises(AttributeError):
            operation.complete = False

    def test_poll_already_complete(self):
        client = _Client()
        operation = self._make_one(self.OPERATION_NAME, client)
        operation._complete = True

        with self.assertRaises(ValueError):
            operation.poll()

    def test_poll_false(self):
        from google.longrunning import operations_pb2

        response_pb = operations_pb2.Operation(done=False)
        client = _Client()
        stub = client._operations_stub
        stub._get_operation_response = response_pb
        operation = self._make_one(self.OPERATION_NAME, client)

        self.assertFalse(operation.poll())

        request_pb = stub._get_operation_requested
        self.assertIsInstance(request_pb, operations_pb2.GetOperationRequest)
        self.assertEqual(request_pb.name, self.OPERATION_NAME)

    def test_poll_true(self):
        from google.longrunning import operations_pb2

        response_pb = operations_pb2.Operation(done=True)
        client = _Client()
        stub = client._operations_stub
        stub._get_operation_response = response_pb
        operation = self._make_one(self.OPERATION_NAME, client)

        self.assertTrue(operation.poll())

        request_pb = stub._get_operation_requested
        self.assertIsInstance(request_pb, operations_pb2.GetOperationRequest)
        self.assertEqual(request_pb.name, self.OPERATION_NAME)

    def test_poll_http(self):
        from google.protobuf.struct_pb2 import Struct
        from google.cloud._testing import _Monkey
        from google.cloud import operation as MUT

        type_url = "type.googleapis.com/%s" % (Struct.DESCRIPTOR.full_name,)
        name = "2302903294023"
        api_response = {
            "name": name,
            "done": True,
            "metadata": {"@type": type_url, "value": {"foo": "Bar"}},
        }
        connection = _Connection(api_response)
        client = _Client(connection)
        operation = self._make_one(name, client)
        operation._from_grpc = False

        with _Monkey(MUT, _TYPE_URL_MAP={type_url: Struct}):
            self.assertTrue(operation.poll())

        expected_path = "operations/%s" % (name,)
        self.assertEqual(
            connection._requested, [{"method": "GET", "path": expected_path}]
        )

    def test__update_state_done(self):
        from google.longrunning import operations_pb2

        operation = self._make_one(None, None)
        self.assertFalse(operation.complete)
        operation_pb = operations_pb2.Operation(done=True)
        operation._update_state(operation_pb)
        self.assertTrue(operation.complete)

    def test__update_state_metadata(self):
        from google.longrunning import operations_pb2
        from google.protobuf.any_pb2 import Any
        from google.protobuf.struct_pb2 import Value
        from google.cloud._testing import _Monkey
        from google.cloud import operation as MUT

        operation = self._make_one(None, None)
        self.assertIsNone(operation.metadata)

        val_pb = Value(number_value=1337)
        type_url = "type.googleapis.com/%s" % (Value.DESCRIPTOR.full_name,)
        val_any = Any(type_url=type_url, value=val_pb.SerializeToString())
        operation_pb = operations_pb2.Operation(metadata=val_any)

        with _Monkey(MUT, _TYPE_URL_MAP={type_url: Value}):
            operation._update_state(operation_pb)

        self.assertEqual(operation.metadata, val_pb)

    def test__update_state_error(self):
        from google.longrunning import operations_pb2
        from google.rpc.status_pb2 import Status

        operation = self._make_one(None, None)
        self.assertIsNone(operation.error)
        self.assertIsNone(operation.response)

        error_pb = Status(code=1)
        operation_pb = operations_pb2.Operation(error=error_pb)
        operation._update_state(operation_pb)

        self.assertEqual(operation.error, error_pb)
        self.assertIsNone(operation.response)

    def test__update_state_response(self):
        from google.longrunning import operations_pb2
        from google.protobuf.any_pb2 import Any
        from google.protobuf.struct_pb2 import Value
        from google.cloud._testing import _Monkey
        from google.cloud import operation as MUT

        operation = self._make_one(None, None)
        self.assertIsNone(operation.error)
        self.assertIsNone(operation.response)

        response_pb = Value(string_value="totes a response")
        type_url = "type.googleapis.com/%s" % (Value.DESCRIPTOR.full_name,)
        response_any = Any(type_url=type_url, value=response_pb.SerializeToString())
        operation_pb = operations_pb2.Operation(response=response_any)

        with _Monkey(MUT, _TYPE_URL_MAP={type_url: Value}):
            operation._update_state(operation_pb)

        self.assertIsNone(operation.error)
        self.assertEqual(operation.response, response_pb)

    def test__update_state_no_result(self):
        from google.longrunning import operations_pb2

        operation = self._make_one(None, None)
        self.assertIsNone(operation.error)
        self.assertIsNone(operation.response)

        operation_pb = operations_pb2.Operation()
        operation._update_state(operation_pb)

        # Make sure nothing changed.
        self.assertIsNone(operation.error)
        self.assertIsNone(operation.response)


class _OperationsStub(object):
    def GetOperation(self, request_pb):
        self._get_operation_requested = request_pb
        return self._get_operation_response


class _Connection(object):
    def __init__(self, *responses):
        self._responses = responses
        self._requested = []

    def api_request(self, **kw):
        self._requested.append(kw)
        response, self._responses = self._responses[0], self._responses[1:]
        return response


class _Client(object):
    def __init__(self, connection=None):
        self._operations_stub = _OperationsStub()
        self._connection = connection



# Copyright 2014 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import requests
from six.moves import http_client

from google.cloud import exceptions


def test_create_google_cloud_error():
    exception = exceptions.GoogleCloudError("Testing")
    exception.code = 600
    assert str(exception) == "600 Testing"
    assert exception.message == "Testing"
    assert exception.errors == []


def test_create_google_cloud_error_with_args():
    error = {
        "domain": "global",
        "location": "test",
        "locationType": "testing",
        "message": "Testing",
        "reason": "test",
    }
    exception = exceptions.GoogleCloudError("Testing", [error])
    exception.code = 600
    assert str(exception) == "600 Testing"
    assert exception.message == "Testing"
    assert exception.errors == [error]


def test_from_http_status():
    message = "message"
    exception = exceptions.from_http_status(http_client.NOT_FOUND, message)
    assert exception.code == http_client.NOT_FOUND
    assert exception.message == message
    assert exception.errors == []


def test_from_http_status_with_errors():
    message = "message"
    errors = ["1", "2"]
    exception = exceptions.from_http_status(
        http_client.NOT_FOUND, message, errors=errors
    )

    assert isinstance(exception, exceptions.NotFound)
    assert exception.code == http_client.NOT_FOUND
    assert exception.message == message
    assert exception.errors == errors


def test_from_http_status_unknown_code():
    message = "message"
    status_code = 156
    exception = exceptions.from_http_status(status_code, message)
    assert exception.code == status_code
    assert exception.message == message


def make_response(content):
    response = requests.Response()
    response._content = content
    response.status_code = http_client.NOT_FOUND
    response.request = requests.Request(
        method="POST", url="https://example.com"
    ).prepare()
    return response


def test_from_http_response_no_content():
    response = make_response(None)

    exception = exceptions.from_http_response(response)

    assert isinstance(exception, exceptions.NotFound)
    assert exception.code == http_client.NOT_FOUND
    assert exception.message == "POST https://example.com/: unknown error"
    assert exception.response == response


def test_from_http_response_text_content():
    response = make_response(b"message")

    exception = exceptions.from_http_response(response)

    assert isinstance(exception, exceptions.NotFound)
    assert exception.code == http_client.NOT_FOUND
    assert exception.message == "POST https://example.com/: message"


def test_from_http_response_json_content():
    response = make_response(
        json.dumps({"error": {"message": "json message", "errors": ["1", "2"]}}).encode(
            "utf-8"
        )
    )

    exception = exceptions.from_http_response(response)

    assert isinstance(exception, exceptions.NotFound)
    assert exception.code == http_client.NOT_FOUND
    assert exception.message == "POST https://example.com/: json message"
    assert exception.errors == ["1", "2"]


def test_from_http_response_bad_json_content():
    response = make_response(json.dumps({"meep": "moop"}).encode("utf-8"))

    exception = exceptions.from_http_response(response)

    assert isinstance(exception, exceptions.NotFound)
    assert exception.code == http_client.NOT_FOUND
    assert exception.message == "POST https://example.com/: unknown error"



# Copyright 2017 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import mock

from google.cloud import obsolete


def test_complain_noop():
    with mock.patch.object(warnings, "warn", autospec=True) as warn:
        obsolete.complain("bogus_package")
        assert warn.call_count == 0


def test_complain():
    with mock.patch.object(warnings, "warn", autospec=True) as warn:
        obsolete.complain("google-cloud-core")
        warn.assert_called_once_with(mock.ANY, DeprecationWarning)



# Copyright 2015 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import unittest

import mock


def _make_credentials():
    import google.auth.credentials

    return mock.Mock(spec=google.auth.credentials.CredentialsWithQuotaProject)


class Test_ClientFactoryMixin(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from google.cloud.client import _ClientFactoryMixin

        return _ClientFactoryMixin

    def test_virtual(self):
        klass = self._get_target_class()
        self.assertFalse("__init__" in klass.__dict__)


class TestClient(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from google.cloud.client import Client

        return Client

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def test_unpickleable(self):
        import pickle

        credentials = _make_credentials()
        HTTP = object()

        client_obj = self._make_one(credentials=credentials, _http=HTTP)
        with self.assertRaises(pickle.PicklingError):
            pickle.dumps(client_obj)

    def test_ctor_defaults(self):
        credentials = _make_credentials()

        patch = mock.patch("google.auth.default", return_value=(credentials, None))
        with patch as default:
            client_obj = self._make_one()

        self.assertIs(client_obj._credentials, credentials)
        self.assertIsNone(client_obj._http_internal)
        default.assert_called_once_with(scopes=None)

    def test_ctor_explicit(self):
        credentials = _make_credentials()
        http = mock.sentinel.http
        client_obj = self._make_one(credentials=credentials, _http=http)

        self.assertIs(client_obj._credentials, credentials)
        self.assertIs(client_obj._http_internal, http)

    def test_ctor_client_options_w_conflicting_creds(self):
        from google.api_core.exceptions import DuplicateCredentialArgs

        credentials = _make_credentials()
        client_options = {"credentials_file": "/path/to/creds.json"}
        with self.assertRaises(DuplicateCredentialArgs):
            self._make_one(credentials=credentials, client_options=client_options)

    def test_ctor_bad_credentials(self):
        credentials = mock.sentinel.credentials

        with self.assertRaises(ValueError):
            self._make_one(credentials=credentials)

    def test_ctor_client_options_w_creds_file_scopes(self):
        credentials = _make_credentials()
        credentials_file = "/path/to/creds.json"
        scopes = ["SCOPE1", "SCOPE2"]
        client_options = {"credentials_file": credentials_file, "scopes": scopes}

        patch = mock.patch(
            "google.auth.load_credentials_from_file", return_value=(credentials, None)
        )
        with patch as load_credentials_from_file:
            client_obj = self._make_one(client_options=client_options)

        self.assertIs(client_obj._credentials, credentials)
        self.assertIsNone(client_obj._http_internal)
        load_credentials_from_file.assert_called_once_with(
            credentials_file, scopes=scopes
        )

    def test_ctor_client_options_w_quota_project(self):
        credentials = _make_credentials()
        quota_project_id = "quota-project-123"
        client_options = {"quota_project_id": quota_project_id}

        client_obj = self._make_one(
            credentials=credentials, client_options=client_options
        )

        self.assertIs(
            client_obj._credentials, credentials.with_quota_project.return_value
        )
        credentials.with_quota_project.assert_called_once_with(quota_project_id)

    def test_ctor__http_property_existing(self):
        credentials = _make_credentials()
        http = object()
        client = self._make_one(credentials=credentials, _http=http)
        self.assertIs(client._http_internal, http)
        self.assertIs(client._http, http)

    def test_ctor__http_property_new(self):
        from google.cloud.client import _CREDENTIALS_REFRESH_TIMEOUT

        credentials = _make_credentials()
        mock_client_cert_source = mock.Mock()
        client_options = {"client_cert_source": mock_client_cert_source}
        client = self._make_one(credentials=credentials, client_options=client_options)
        self.assertIsNone(client._http_internal)

        with mock.patch(
            "google.auth.transport.requests.AuthorizedSession"
        ) as AuthorizedSession:
            session = mock.Mock()
            session.configure_mtls_channel = mock.Mock()
            AuthorizedSession.return_value = session
            self.assertIs(client._http, session)
            # Check the mock.
            AuthorizedSession.assert_called_once_with(
                credentials, refresh_timeout=_CREDENTIALS_REFRESH_TIMEOUT
            )
            session.configure_mtls_channel.assert_called_once_with(
                mock_client_cert_source
            )
            # Make sure the cached value is used on subsequent access.
            self.assertIs(client._http_internal, session)
            self.assertIs(client._http, session)
            self.assertEqual(AuthorizedSession.call_count, 1)

    def test_from_service_account_info(self):
        klass = self._get_target_class()

        info = {"dummy": "value", "valid": "json"}
        constructor_patch = mock.patch(
            "google.oauth2.service_account.Credentials.from_service_account_info",
            return_value=_make_credentials(),
        )

        with constructor_patch as constructor:
            client_obj = klass.from_service_account_info(info)

        self.assertIs(client_obj._credentials, constructor.return_value)
        self.assertIsNone(client_obj._http_internal)
        constructor.assert_called_once_with(info)

    def test_from_service_account_info_w_explicit_credentials(self):
        KLASS = self._get_target_class()

        info = {"dummy": "value", "valid": "json"}

        with self.assertRaises(TypeError):
            KLASS.from_service_account_info(info, credentials=mock.sentinel.credentials)

    def test_from_service_account_json(self):
        from google.cloud import _helpers

        klass = self._get_target_class()

        info = {"dummy": "value", "valid": "json"}
        json_file = io.StringIO(_helpers._bytes_to_unicode(json.dumps(info)))

        file_open_patch = mock.patch("io.open", return_value=json_file)
        constructor_patch = mock.patch(
            "google.oauth2.service_account.Credentials." "from_service_account_info",
            return_value=_make_credentials(),
        )

        with file_open_patch as file_open:
            with constructor_patch as constructor:
                client_obj = klass.from_service_account_json(mock.sentinel.filename)

        self.assertIs(client_obj._credentials, constructor.return_value)
        self.assertIsNone(client_obj._http_internal)
        # Check that mocks were called as expected.
        file_open.assert_called_once_with(mock.sentinel.filename, "r", encoding="utf-8")
        constructor.assert_called_once_with(info)

    def test_close_w__http_internal_none(self):
        credentials = _make_credentials()
        client_obj = self._make_one(credentials=credentials, _http=None)

        client_obj.close()  # noraise

    def test_close_w__http_internal_set(self):
        credentials = _make_credentials()
        http = mock.Mock(spec=["close"])
        client_obj = self._make_one(credentials=credentials, _http=http)

        client_obj.close()

        http.close.assert_called_once_with()


class Test_ClientProjectMixin(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from google.cloud.client import _ClientProjectMixin

        return _ClientProjectMixin

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def test_ctor_defaults_wo_envvar(self):
        environ = {}
        patch_env = mock.patch("os.environ", new=environ)
        patch_default = mock.patch(
            "google.cloud.client._determine_default_project", return_value=None,
        )
        with patch_env:
            with patch_default as patched:
                with self.assertRaises(EnvironmentError):
                    self._make_one()

        patched.assert_called_once_with(None)

    def test_ctor_defaults_w_envvar(self):
        from google.auth.environment_vars import PROJECT

        project = "some-project-123"
        environ = {PROJECT: project}
        patch_env = mock.patch("os.environ", new=environ)
        with patch_env:
            client = self._make_one()

        self.assertEqual(client.project, project)

    def test_ctor_defaults_w_legacy_envvar(self):
        from google.auth.environment_vars import LEGACY_PROJECT

        project = "some-project-123"
        environ = {LEGACY_PROJECT: project}
        patch_env = mock.patch("os.environ", new=environ)
        with patch_env:
            client = self._make_one()

        self.assertEqual(client.project, project)

    def test_ctor_w_explicit_project(self):
        explicit_project = "explicit-project-456"
        patch_default = mock.patch(
            "google.cloud.client._determine_default_project", return_value=None,
        )
        with patch_default as patched:
            client = self._make_one(project=explicit_project)

        self.assertEqual(client.project, explicit_project)

        patched.assert_not_called()

    def test_ctor_w_explicit_project_bytes(self):
        explicit_project = b"explicit-project-456"
        patch_default = mock.patch(
            "google.cloud.client._determine_default_project", return_value=None,
        )
        with patch_default as patched:
            client = self._make_one(project=explicit_project)

        self.assertEqual(client.project, explicit_project.decode("utf-8"))

        patched.assert_not_called()

    def test_ctor_w_explicit_project_invalid(self):
        explicit_project = object()
        patch_default = mock.patch(
            "google.cloud.client._determine_default_project", return_value=None,
        )
        with patch_default as patched:
            with self.assertRaises(ValueError):
                self._make_one(project=explicit_project)

        patched.assert_not_called()

    @staticmethod
    def _make_credentials(**kw):
        from google.auth.credentials import Credentials

        class _Credentials(Credentials):
            def __init__(self, **kw):
                self.__dict__.update(kw)

            def refresh(self):  # pragma: NO COVER
                pass

        return _Credentials(**kw)

    def test_ctor_w_explicit_credentials_wo_project(self):
        default_project = "default-project-123"
        credentials = self._make_credentials()
        patch_default = mock.patch(
            "google.cloud.client._determine_default_project",
            return_value=default_project,
        )
        with patch_default as patched:
            client = self._make_one(credentials=credentials)

        self.assertEqual(client.project, default_project)

        patched.assert_called_once_with(None)

    def test_ctor_w_explicit_credentials_w_project(self):
        project = "credentials-project-456"
        credentials = self._make_credentials(project_id=project)
        patch_default = mock.patch(
            "google.cloud.client._determine_default_project", return_value=None,
        )
        with patch_default as patched:
            client = self._make_one(credentials=credentials)

        self.assertEqual(client.project, project)

        patched.assert_not_called()


class TestClientWithProject(unittest.TestCase):
    @staticmethod
    def _get_target_class():
        from google.cloud.client import ClientWithProject

        return ClientWithProject

    def _make_one(self, *args, **kw):
        return self._get_target_class()(*args, **kw)

    def test_constructor_defaults(self):
        credentials = _make_credentials()
        patch1 = mock.patch("google.auth.default", return_value=(credentials, None))

        project = "prahj-ekt"
        patch2 = mock.patch(
            "google.cloud.client._determine_default_project", return_value=project
        )

        with patch1 as default:
            with patch2 as _determine_default_project:
                client_obj = self._make_one()

        self.assertEqual(client_obj.project, project)
        self.assertIs(client_obj._credentials, credentials)
        self.assertIsNone(client_obj._http_internal)
        default.assert_called_once_with(scopes=None)
        _determine_default_project.assert_called_once_with(None)

    def test_constructor_missing_project(self):
        from google.cloud._testing import _Monkey
        from google.cloud import client

        FUNC_CALLS = []

        def mock_determine_proj(project):
            FUNC_CALLS.append((project, "_determine_default_project"))
            return None

        with _Monkey(client, _determine_default_project=mock_determine_proj):
            self.assertRaises(EnvironmentError, self._make_one)

        self.assertEqual(FUNC_CALLS, [(None, "_determine_default_project")])

    def test_constructor_w_invalid_project(self):
        CREDENTIALS = _make_credentials()
        HTTP = object()
        with self.assertRaises(ValueError):
            self._make_one(project=object(), credentials=CREDENTIALS, _http=HTTP)

    def _explicit_ctor_helper(self, project):
        import six

        CREDENTIALS = _make_credentials()
        HTTP = object()

        client_obj = self._make_one(
            project=project, credentials=CREDENTIALS, _http=HTTP
        )

        if isinstance(project, six.binary_type):
            self.assertEqual(client_obj.project, project.decode("utf-8"))
        else:
            self.assertEqual(client_obj.project, project)
        self.assertIs(client_obj._credentials, CREDENTIALS)
        self.assertIs(client_obj._http_internal, HTTP)

    def test_constructor_explicit_bytes(self):
        PROJECT = b"PROJECT"
        self._explicit_ctor_helper(PROJECT)

    def test_constructor_explicit_unicode(self):
        PROJECT = u"PROJECT"
        self._explicit_ctor_helper(PROJECT)

    def _from_service_account_info_helper(self, project=None):
        klass = self._get_target_class()

        default_project = "eye-d-of-project"
        info = {"dummy": "value", "valid": "json", "project_id": default_project}
        kwargs = {}

        if project is None:
            expected_project = default_project
        else:
            expected_project = project
            kwargs["project"] = project

        constructor_patch = mock.patch(
            "google.oauth2.service_account.Credentials.from_service_account_info",
            return_value=_make_credentials(),
        )

        with constructor_patch as constructor:
            client_obj = klass.from_service_account_info(info, **kwargs)

        self.assertIs(client_obj._credentials, constructor.return_value)
        self.assertIsNone(client_obj._http_internal)
        self.assertEqual(client_obj.project, expected_project)

        constructor.assert_called_once_with(info)

    def test_from_service_account_info(self):
        self._from_service_account_info_helper()

    def test_from_service_account_info_with_project(self):
        self._from_service_account_info_helper(project="prah-jekt")

    def test_from_service_account_info_with_posarg(self):
        class Derived(self._get_target_class()):
            def __init__(self, required, **kwargs):
                super(Derived, self).__init__(**kwargs)
                self.required = required

        project = "eye-d-of-project"
        info = {"dummy": "value", "valid": "json", "project_id": project}

        # Mock both the file opening and the credentials constructor.
        constructor_patch = mock.patch(
            "google.oauth2.service_account.Credentials.from_service_account_info",
            return_value=_make_credentials(),
        )

        with constructor_patch as constructor:
            client_obj = Derived.from_service_account_info(info, "REQUIRED")

        self.assertIsInstance(client_obj, Derived)
        self.assertIs(client_obj._credentials, constructor.return_value)
        self.assertIsNone(client_obj._http_internal)
        self.assertEqual(client_obj.project, project)
        self.assertEqual(client_obj.required, "REQUIRED")

        # Check that mocks were called as expected.
        constructor.assert_called_once_with(info)

    def _from_service_account_json_helper(self, project=None):
        from google.cloud import _helpers

        klass = self._get_target_class()

        default_project = "eye-d-of-project"
        info = {"dummy": "value", "valid": "json", "project_id": default_project}
        if project is None:
            expected_project = "eye-d-of-project"
        else:
            expected_project = project

        # Mock both the file opening and the credentials constructor.
        json_fi = io.StringIO(_helpers._bytes_to_unicode(json.dumps(info)))
        file_open_patch = mock.patch("io.open", return_value=json_fi)
        constructor_patch = mock.patch(
            "google.oauth2.service_account.Credentials." "from_service_account_info",
            return_value=_make_credentials(),
        )

        with file_open_patch as file_open:
            with constructor_patch as constructor:
                kwargs = {}
                if project is not None:
                    kwargs["project"] = project
                client_obj = klass.from_service_account_json(
                    mock.sentinel.filename, **kwargs
                )

        self.assertIs(client_obj._credentials, constructor.return_value)
        self.assertIsNone(client_obj._http_internal)
        self.assertEqual(client_obj.project, expected_project)
        # Check that mocks were called as expected.
        file_open.assert_called_once_with(mock.sentinel.filename, "r", encoding="utf-8")
        constructor.assert_called_once_with(info)

    def test_from_service_account_json(self):
        self._from_service_account_json_helper()

    def test_from_service_account_json_project_set(self):
        self._from_service_account_json_helper(project="prah-jekt")

    def test_from_service_account_json_with_posarg(self):
        from google.cloud import _helpers

        class Derived(self._get_target_class()):
            def __init__(self, required, **kwargs):
                super(Derived, self).__init__(**kwargs)
                self.required = required

        project = "eye-d-of-project"
        info = {"dummy": "value", "valid": "json", "project_id": project}

        # Mock both the file opening and the credentials constructor.
        json_fi = io.StringIO(_helpers._bytes_to_unicode(json.dumps(info)))
        file_open_patch = mock.patch("io.open", return_value=json_fi)
        constructor_patch = mock.patch(
            "google.oauth2.service_account.Credentials.from_service_account_info",
            return_value=_make_credentials(),
        )

        with file_open_patch as file_open:
            with constructor_patch as constructor:
                client_obj = Derived.from_service_account_json(
                    mock.sentinel.filename, "REQUIRED"
                )

        self.assertIsInstance(client_obj, Derived)
        self.assertIs(client_obj._credentials, constructor.return_value)
        self.assertIsNone(client_obj._http_internal)
        self.assertEqual(client_obj.project, project)
        self.assertEqual(client_obj.required, "REQUIRED")

        # Check that mocks were called as expected.
        file_open.assert_called_once_with(mock.sentinel.filename, "r", encoding="utf-8")
        constructor.assert_called_once_with(info)
